import os
import sqlalchemy as sa
from . import TimeHelper
from .app import config
from .connections import ConnectionManager
from .models import (
    automated_reports,
    automated_report_contacts,
    automated_report_recipients,
)
from .mailer import Mailer
from .related_record import RelatedRecord
from .queries import QueryGenerator
from .xlsx import WorkbookBuilder
from .logger import PortholeLogger


class ReportWriter:
    """
    The purpose of this class is to use the QueryGenerator and WorkbookBuilder
    together to make an Excel file and populate it with data.
    """
    def __init__(self, report_title, logger=None):
        self.report_title = report_title
        self.report_file = None
        self.file_path = config['Default'].get('base_file_path')
        self.workbook_builder = None
        self.record_count = 0
        self.logger = logger or PortholeLogger(report_title)

    def build_file(self):
        """
        Loads file and path information and creates XLSX Writer Workbook object.
        Default file is named with convention yyyy-mm-dd - Report Name.xlsx
        """
        try:
            # Construct the file path and create the workbook.
            local_timezone = config['Default'].get('local_timezone') or 'UTC'
            today = TimeHelper.today(timezone=local_timezone)
            report_file_name = f"{today} - {self.report_title}.xlsx"
            self.report_file = os.path.join(self.file_path, report_file_name)
            self.workbook_builder = WorkbookBuilder(filename=self.report_file)
        except:
            error = "Unable to build file for {}".format(self.report_title)
            self.logger.exception(error)

    def close_workbook(self):
        """
        Closes workbook object after creation is complete.
        Should always be executed before sending file.
        """
        if self.workbook_builder:
            self.workbook_builder.workbook.close()

    def add_format(self, format_name, format_params):
        self.workbook_builder.add_format(format_name, format_params)

    def execute_query(self, cm, query=None, sql=None, increment_counter=True):
        """
        Args:
            cm              (ConnectionManager):
                            Object which contains connection to desired database.
            sheet_name      (str): The name of the worksheet to be created.
            query           (dict): Optional. May contain filename and params for a query
                                to be executed. If included, filename is required.
                    filename      (str): Name of file to be read without extension.
                    params        (dict): Optional. Contains parameter names and values,
                                if applicable. Should only be included along with
                                query_file containing parameter placeholders.
            sql             (str or sqlalchemy.sql.selectable.Select statement):
                                Optional. A SQL query ready for execution.

        Executes SQL and returns QueryResult object, containing data and metadata.
        """
        if query is None:
            query = {}
        filename = query.get('filename')
        params = query.get('params')
        q = QueryGenerator(cm=cm, filename=filename, params=params, sql=sql, logger=self.logger)
        try:
            results = q.execute()
            if increment_counter:
                self.record_count += results.result_count
            return results
        except:
            error = "Unable to execute query {}".format(query.get('filename'))
            self.logger.exception(error)

    def make_worksheet(self, sheet_name, query_results, **kwargs):
        """Adds worksheet to workbook using provided query results."""
        try:
            self.workbook_builder.add_worksheet(
                sheet_name=sheet_name,
                field_names=query_results.field_names,
                sheet_data=query_results.result_data,
                **kwargs
            )
        except:
            error = "Unable to add worksheet {}".format(sheet_name)
            self.logger.exception(error)

    def create_worksheet_from_query(self, cm, sheet_name, query=None, sql=None, query_kwargs=None, worksheet_kwargs=None):
        """
        Args:
            cm              (ConnectionManager):
                            Object which contains connection to desired database.
            sheet_name      (str): The name of the worksheet to be created.
            query           (dict): Optional. May contain filename and params for a query
                                to be executed. If included, filename is required.
                filename      (str): Name of file to be read without extension.
                params        (dict): Optional. Contains parameter names and values,
                                if applicable. Should only be included along with
                                query_file containing parameter placeholders.
            sql             (str or sqlalchemy.sql.selectable.Select statement):
                                Optional. A SQL query ready for execution.
            query_kwargs    (dict): Optional. Dictionary of keyword arguments to pass to `execute_query`.
            worksheet_kwargs (dict): Optional. Dictionary of keyword arguments to pass to `make_worksheet`

        Executes a query and uses results to add worksheet to ReportWriter.workbook_builder.
        """
        if query is None:
            query = {}
        if query_kwargs is None:
            query_kwargs = {}
        if worksheet_kwargs is None:
            worksheet_kwargs = {}
        results = self.execute_query(cm=cm, query=query, sql=sql, **query_kwargs)
        self.make_worksheet(sheet_name=sheet_name, query_results=results, **worksheet_kwargs)


class DatabaseLogger:
    """
    Don't want to log if report not active,
    or if logging is not enabled.
    """
    def __init__(self, cm, report_name, log_table='report_logs', logger=None):
        self.cm = cm
        self.report_name = report_name
        self.log_table = log_table
        self.logger = logger or PortholeLogger(report_name)
        self.report_log = None

    def create_record(self):
        """
        Inserts new log record into default report logging table
        using RelatedRecord and saves instance of RelatedRecord for updating
        later on with results of report execution.
        """
        report_log = RelatedRecord(self.cm, self.log_table)
        try:
            report_log.insert({
                'report_name': self.report_name,
                'started_at': TimeHelper.now(string=False)
            })
        except:
            self.logger.exception("Unable to create log record.")
        self.report_log = report_log
        self.logger.extra.update(report_log_id=report_log.primary_key)

    def finalize_record(self):
        """Update log at conclusion of report execution to indicate success/failure."""
        error_buffer = self.logger.error_buffer.buffer
        if error_buffer:
            data_to_update = {
                'completed_at': TimeHelper.now(string=False),
                'success': 0,
                'error_detail': "; ".join([str(log.exc_text) for log in error_buffer if hasattr(log, 'exc_text')])[:255]
            }
        else:
            data_to_update = {
                'completed_at': TimeHelper.now(string=False),
                'success': 1
            }
        try:
            self.report_log.update(data_to_update)
        except:
            self.logger.exception("Unable to finalize log record.")


class NewDBLogger:
    def __init__(
            self,
            cm: ConnectionManager,
            report_name: str, 
            log_table: sa.Table, 
            logger: PortholeLogger = None
    ):
        self.cm = cm
        self.report_name = report_name
        self.log_table = log_table
        self.logger = logger or PortholeLogger(report_name)

    def create(self):
        statement = self.log_table.insert().values({
            "report_name": self.report_name,
            "started_at": TimeHelper.now(string=False)
        })
        try:
            self.cm.conn.execute(statement)
        except:
            self.logger.exception("Unable to create log record.")


class ReportActiveChecker:

    def __init__(self, cm, report_name, logger=None):
        self.cm = cm
        self.report_name = report_name
        self.active = False
        self.logger = logger or PortholeLogger(report_name)
        self.check_if_active()

    def __bool__(self):
        return self.active

    def check_if_active(self):
        """Queries database to determine if the 'active' attribute for this report
        is set to 1 (active) or 0 (inactive).
        """
        statement = sa.select(
            [automated_reports.c.active]
        ).where(
            automated_reports.c.report_name == self.report_name
            )
        try:
            q = QueryGenerator(cm=self.cm, sql=statement, logger=self.logger)
            results = q.execute()
            if results.result_data[0]['active'] == 1:
                self.active = True
            else:
                self.active = False
        except IndexError:
            error = (
                f"Report <{self.report_name}> was not found in the {automated_reports.name} table. "
                f"Ensure <{self.report_name}> matches a record in the table and try again."
            )
            self.logger.exception(error)
            raise Exception(error)


class RecipientsChecker:
    def __init__(self, cm, report_name, logger=None):
        self.cm = cm
        self.report_name = report_name
        self.to_recipients = []
        self.cc_recipients = []
        self.logger = logger or PortholeLogger(report_name)

    def get_recipients(self):
        """Performs lookup in database for report recipients based on report name."""
        statement = sa.select([
            automated_report_contacts.c.email_address,
            automated_report_recipients.c.recipient_type
        ]).select_from(
            automated_reports.join(automated_report_recipients).join(automated_report_contacts)).where(
                        automated_reports.c.report_name == self.report_name
        )
        try:
            q = QueryGenerator(cm=self.cm, sql=statement, logger=self.logger)
            results = q.execute()
            for recipient in results.row_proxies:
                if recipient.recipient_type == 'to':
                    self.to_recipients.append(recipient.email_address)
                else:
                    self.cc_recipients.append(recipient.email_address)
        except:
            error = "Error getting recipients for {}".format(self.report_name)
            self.logger.exception(error)
        if not self.to_recipients:
            error = "No primary recipient found for {}".format(self.report_name)
            self.logger.error(error)
            raise KeyError("No primary recipient found for {}".format(self.report_name))
        return self.to_recipients, self.cc_recipients


class ReportErrorNotifier:

    def __init__(self, report_title, error_buffer):
        self.report_title = report_title
        self.notification_recipient = config['Default'].get('notification_recipient')
        self.error_buffer = error_buffer
        self.notified = False

    def send_log_by_email(self):
        email = Mailer()
        email.recipients = [self.notification_recipient]
        email.cc_recipients = ['']
        email.subject = "Report Failure: {}".format(self.report_title)
        email.message = self.construct_message()
        email.attachments = []
        email.send_email()
        self.notified = True

    def construct_message(self):
        msg = """Execution of the following report failed: {}

The following errors were logged:\n""".format(self.report_title)
        msg = msg.format(self.report_title)
        for i, error in enumerate(self.error_buffer, 1):
            msg += str(i) + '. ' + str(error) + '\n'
        return msg



from .app import config
from sqlalchemy import MetaData, ForeignKey, Table, Column, func
from sqlalchemy import Boolean, DateTime, Integer, String, Text

try:
    schema = config[config['Default']['database']]['schema']
except:
    schema = None

metadata = MetaData(schema=schema)

automated_reports = Table(
    'automated_reports',
    metadata,
    Column('report_id', Integer, primary_key=True),
    Column('report_name', String(64), nullable=False),
    Column('active', Integer),
    Column('created_at', DateTime, server_default=func.now()),
    Column('updated_at', DateTime, onupdate=func.now()),
    Column('report_description', String(255)),
    Column('schedule_minute', String(32)),
    Column('schedule_hour', String(32)),
    Column('schedule_day_of_week', String(32)),
    Column('schedule_day_of_month', String(32)),
    Column('schedule_month_of_year', String(32))
)

automated_report_contacts = Table(
    'automated_report_contacts',
    metadata,
    Column('contact_id', Integer, primary_key=True),
    Column('last_name', String(64)),
    Column('first_name', String(64)),
    Column('email_address', String(64)),
    Column('created_at', DateTime, server_default=func.now()),
    Column('updated_at', DateTime, onupdate=func.now()),
)

automated_report_recipients = Table(
    'automated_report_recipients',
    metadata,
    Column('recipient_id', Integer, primary_key=True),
    Column('report_id', Integer, ForeignKey(automated_reports.c.report_id), nullable=False),
    Column('contact_id', Integer, ForeignKey(automated_report_contacts.c.contact_id), nullable=False),
    Column('recipient_type', String(10), nullable=False),
    Column('created_at', DateTime, server_default=func.now()),
    Column('updated_at', DateTime, onupdate=func.now()),
)

report_logs = Table(
    'report_logs', metadata,
    Column('id', Integer, primary_key=True),
    Column('report_name', String(64), nullable=False),
    Column('started_at', DateTime),
    Column('completed_at', DateTime),
    Column('success', Boolean),
    Column('error_detail', String(255)),
    Column('created_at', DateTime, server_default=func.now()),
    Column('updated_at', DateTime, onupdate=func.now()),
)

report_log_details = Table(
    'report_log_details',
    metadata,
    Column('id', Integer, primary_key=True),
    Column('report_log_id', Integer, ForeignKey(report_logs.c.id)),
    Column('level_number', Integer),
    Column('level_name', String(64)),
    Column('logger', String(64)),
    Column('msg', Text),
    Column('traceback', Text),
    Column('created_at', DateTime, server_default=func.now()),
    Column('updated_at', DateTime, onupdate=func.now()),
)



import logging
import traceback
import warnings
from logging.handlers import TimedRotatingFileHandler
import sqlalchemy as sa
from .app import config
from .models import report_log_details


class PortholeLogger(object):
    """
    Logger class. Log to console by default with optional
    logging to file.
    Set behavior in config.ini:
        log_to_file (bool): Whether to write log to file.
        logfile (str):      Name of file to write to.
        logging_db (str):   Name of database connection to write to, if applicable.
    """

    DEFAULT_FORMAT = '%(levelname)s -- %(asctime)s -- %(name)s -- %(message)s'
    DEFAULT_DATE_FORMAT = '%Y-%m-%d %H:%M:%S'

    def __init__(
            self,
            name: str,
            logfile: str = None,
            log_to_db: bool = False,
            log_table: sa.table = report_log_details,
            fmt: str = DEFAULT_FORMAT,
            datefmt: str = DEFAULT_DATE_FORMAT
    ):
        self.log_format = fmt
        self.date_format = datefmt
        self.formatter = None
        self.error_buffer = None
        self.extra = {}
        self.log_to_file = config['Logging'].getboolean('log_to_file', False)
        self.logfile = logfile or config['Logging'].get('logfile', None)
        self.log_to_db = log_to_db
        self.log_table = log_table
        self.rotate_logs = config['Logging'].getboolean('rotate_logs', False)

        self.logger = logging.getLogger(name)
        self.logger.setLevel(logging.INFO)
        self.create_formatter()
        self.setup_handlers()

    def setup_handlers(self):
        self.add_stream_handler()
        self.error_buffer = ErrorBuffer()
        self._add_handler(self.error_buffer)
        if self.log_to_file:
            self.add_file_handler()
        if self.log_to_db:
            self.add_db_handler()

    def create_formatter(self):
        self.formatter = logging.Formatter(
            fmt=self.log_format,
            datefmt=self.date_format
        )

    def _add_handler(self, handler):
        """Ensure handlers are unique by name"""
        handlers = [h.name for h in self.logger.handlers if h.name]
        if handler.name not in handlers:
            self.logger.addHandler(handler)

    def add_stream_handler(self):
        stream_handler = logging.StreamHandler()
        stream_handler.set_name(name="stream_handler")
        stream_handler.setFormatter(self.formatter)
        self._add_handler(stream_handler)

    def add_file_handler(self):
        if self.rotate_logs:
            self.add_rotating_file_handler()
        else:
            handler = logging.FileHandler(self.logfile)
            handler.set_name(name="file_handler")
            handler.setFormatter(self.formatter)
            self._add_handler(handler)

    def add_db_handler(self):
        log_db = config['Logging'].get('logging_db')
        if not log_db:
            self.warning("log_to_db is set to true, but logging_db is not set. Will not log to database.")
            return
        handler = DatabaseHandler(logger=self, database=log_db, table=self.log_table)
        self._add_handler(handler)

    def add_rotating_file_handler(self):
        interval_type = config['Logging.get'].get('rotation_interval_type', 'h')
        interval_magnitude = config['Logging'].get('rotation_interval_magnitude', 1)
        backup_count = config['Logging'].get('backup_count', 0)
        handler = TimedRotatingFileHandler(
            filename=self.logfile,
            when=interval_type,
            interval=interval_magnitude,
            backupCount=backup_count
        )
        handler.set_name(name="rotating_file_handler")
        self._add_handler(handler)

    def debug(self, msg, *args, **kwargs):
        self.logger.debug(msg, *args, **kwargs)

    def info(self, msg, *args, **kwargs):
        self.logger.info(msg, *args, extra=self.extra, **kwargs)

    def warning(self, msg, *args, **kwargs):
        self.logger.warning(msg, *args, extra=self.extra, **kwargs)

    def error(self, msg, *args, **kwargs):
        self.logger.error(msg, *args, extra=self.extra, **kwargs)

    def critical(self, msg, *args, **kwargs):
        self.logger.critical(msg, *args, extra=self.extra, **kwargs)

    def exception(self, msg, *args, **kwargs):
        self.logger.exception(msg, *args, extra=self.extra, **kwargs)

    def set_level(self, level):
        self.logger.setLevel(level)


class DatabaseHandler(logging.Handler):
    def __init__(self, logger, database=None, table: sa.Table = None):
        from .connections import ConnectionManager
        logging.Handler.__init__(self)
        self.set_name(name="database_handler")
        self.database = database
        self.table = table
        self.cm = ConnectionManager(self.database, logger=logger)

    def emit(self, record):
        if record.exc_info:
            trace = traceback.format_exc()
        else:
            trace = None
        data = {
            'report_log_id': record.__dict__.get('report_log_id'),
            'level_number': record.levelno,
            'level_name': record.levelname,
            'msg': record.msg,
            'logger': record.name,
            'traceback': trace
        }
        self._log_record_to_db(data)

    def _log_record_to_db(self, log_data: dict):
        statement = self.table.insert().values(**log_data)
        try:
            self.cm.connect()
            self.cm.conn.execute(statement)
        except Exception as e:
            warnings.warn(f"Exception when writing log to database: {e}")
        finally:
            self.cm.close()


class ErrorBuffer(logging.handlers.BufferingHandler):
    def __init__(self, capacity: int = 1024):
        super().__init__(capacity=capacity)
        self.setLevel(logging.ERROR)
        self.set_name(name="error_buffer")

    @property
    def empty(self):
        return len(self.buffer) == 0

    @property
    def present(self):
        return not self.empty



from sqlalchemy import Table, MetaData


class RelatedRecord(object):
    """
    Represents a database record; provides insert/update functionality of
    of individual records.
    """
    def __init__(self, connection_manager, table):
        self.cm = connection_manager
        self.table = Table(
            table,
            MetaData(schema=connection_manager.schema),
            autoload=True,
            autoload_with=connection_manager.engine
        )
        self.primary_key_name = self.inspect_primary_key()
        self.primary_key = None
        self.inserted = False
        self.updated = False

    def inspect_primary_key(self):
        keys = [column.name for column in self.table.columns.values() if column.primary_key]
        if len(keys) == 1:
            return keys[0]
        else:
            raise ValueError("Expected one primary key, found {}".format(len(keys)))

    def insert(self, data_to_insert):
        if not self.inserted:
            try:
                statement = self.table.insert().values(**data_to_insert)
                result = self.cm.conn.execute(statement)
                self.primary_key = result.inserted_primary_key[0]
                self.inserted = True
            except:
                raise
        else:
            raise RuntimeError("Cannot perform insert - record already inserted.")

    def update(self, date_to_update):
        try:
            key_column = getattr(self.table.c, self.primary_key_name)
            statement = self.table.update().where(
                key_column == self.primary_key
            ).values(**date_to_update)
            self.cm.conn.execute(statement)
            self.updated = True
        except:
            raise


if __name__ == '__main__':
    pass



import os, re, json
from collections import OrderedDict
from decimal import Decimal
from datetime import date
from .app import config
from .connections import ConnectionManager
from .logger import PortholeLogger

RE_SQL_STATEMENT = re.compile(''';(?=(?:[^"'`]*["'`][^"'`]*["'`])*[^"'`]*$)''')


class QueryResult(object):
    """Represent result data from an executed query. Includes capability to write results as json."""

    def __init__(self, result_count=None, field_names=None, result_data=None, row_proxies=None):
        self.result_count = result_count
        self.field_names = field_names
        self.result_data = [RowDict(fields=field_names, values=row) for row in result_data]
        self.row_proxies = row_proxies
        self.field_index = {field: idx for idx, field in enumerate(field_names)}

    @staticmethod
    def json_converter(obj):
        """Required to convert datatypes not otherwise json serializable."""
        if isinstance(obj, Decimal):
            return float(obj)
        elif isinstance(obj, date):
            return obj.isoformat()
        else:
            raise TypeError("Cannot convert provided type {}".format(type(obj)))

    def as_dict(self):
        raise DeprecationWarning("QueryResult.as_dict method is no longer available and will be removed.")

    def write_to_json(self, filename):
        with open(filename, 'w') as f:
            json.dump(self.result_data, f, default=self.json_converter)

    def map_function_to_field(self, field, func):
        assert field in self.field_names
        for row in self.result_data:
            row[field] = func(row[field])

    def apply(self, func):
        for row in self.result_data:
            func(row)


class RowDict(OrderedDict):
    """
    RowDict is used to represent a record in a query result. Because RowDict inherits from
    OrderedDict, values are accessible using keys (field/header names). RowDict has special
    iteration behavior such that iterating over a RowDict instance will return values rather
    than keys.
    RowDict inherits from OrderedDict to preserve compatibility with Python < 3.6.
    """
    def __init__(self, data=None, fields=None, values=None):
        fields_and_values_provided = fields is not None and values is not None
        if fields_and_values_provided is True and data is None:
            if len(fields) > len(set(fields)):
                raise ValueError("Field names must be unique, but your result set contains non-unique field names.")
            data = OrderedDict(zip(fields, values))
        else:
            data = data or OrderedDict()
        super().__init__(data)

    def __iter__(self):
        return iter(self.values())

    def __repr__(self):
        if not self:
            return '%s()' % (self.__class__.__name__,)
        return '%s(%r)' % (self.__class__.__name__, list(self.items()))


class QueryGenerator(object):
    """Execute SQL query and return results"""
    def __init__(
            self,
            cm,
            filepath=None,
            filename=None,
            params=None,
            sql=None,
            multiple_statements=False,
            logger=None
    ):
        if filename is not None and sql is not None:
            raise TypeError("Cannot give both 'filename' and 'sql' arguments")
        self.logger = logger or PortholeLogger(name=__name__)
        self.cm = cm
        self.filepath = filepath
        self.filename = filename
        self.params = params
        self.raw_sql = sql
        self.sql = None
        self.multiple_statements = multiple_statements

    def construct_query(self):
        """Read and parameterize (if necessary) a .sql file for execution."""
        reader = QueryReader(filepath=self.filepath, filename=self.filename, raw_sql=self.raw_sql, params=self.params)
        return reader.sql

    def execute(self):
        """
        This method will execute a series of statements, if that is what has been provided.
        The first returnable set of data will be returned - if the statements provided
        include multiple selects, results from only the first will be returned.
        Future implementations may allow for a sequence of results to be returned.
        """
        if self.sql is None:
            self.sql = self.construct_query()
        # Only SQL strings can be split, not (e.g.) SQLAlchemy statements.
        if self.multiple_statements and isinstance(self.sql, str):
            statements = self._split_sql()
        else:
            statements = [self.sql]
        single_statement = True if len(statements) == 1 and self.filename else False
        try:
            for statement in statements:
                result_proxy = self.cm.conn.execute(statement)
                log_string = self.filename if single_statement else str(statement)[:25]
                self.logger.info("Executed {} against {}".format(log_string, self.cm.db))
            if result_proxy.cursor:
                return self.fetch_results(result_proxy)
        except Exception as e:
            self.logger.exception(e)
            raise

    def _split_sql(self):
        """
        Returns a list containing individual sql statements as strings to be executed.
        """
        return [stmt.strip() for stmt in RE_SQL_STATEMENT.split(self.sql) if stmt.strip()]

    @staticmethod
    def fetch_results(result_proxy):
        field_names = result_proxy.keys()
        row_proxies = result_proxy.fetchall()
        result_data = [row.values() for row in row_proxies]
        query_results = QueryResult(result_count=len(result_data),
                                    field_names=field_names,
                                    result_data=result_data,
                                    row_proxies=row_proxies)
        return query_results


class QueryReader(object):
    """
    QueryReader is used to read, and optionally to parameterize, .sql files.

    By default, it is assumed that your query lives in your project's default
    query path. Override this behavior by providing a value for the `filepath`
    parameter.

    Saved queries requiring parameters at runtime should use placeholders in the format
    matching the raw_pattern attribute, which is by default:
        #{parameter_name} e.g. #{first_name}, #{last_name}, etc.

    Replacement values must be provided for all parameter placeholders.

    Simple Usage - No Parameters

    >> my_query = QueryReader(filename='my_query')
    >> my_query.sql
    select * from table;

    Usage with Parameters

    >> my_query = QueryReader(filename='my_query', params={'field': 'value'})
    >> my_query.sql
    select * from table where field = 'value';

    """
    def __init__(self, filepath=None, filename=None, raw_sql=None, params=None):
        self.filename = filename
        self.params = params
        self.raw_sql = raw_sql
        self.sql = None
        self.to_replace = None
        self.query_path = filepath or config['Default']['query_path']
        self.raw_pattern = '(#{[a-zA-Z_]*})'
        self.process_sql()

    def __repr__(self):
        if self.sql:
            return self.sql
        elif self.raw_sql:
            return self.raw_sql
        else:
            return "< Empty QueryReader object >"

    def process_sql(self):
        if self.filename:
            self.read()
        if isinstance(self.raw_sql, str):
            self.find_values_to_replace()
        if self.to_replace:
            self.replace_params()
            self.validate()
        else:
            self.sql = self.raw_sql

    def read(self):
        """Reads and stores query contents"""
        file_path = os.path.join(self.query_path, self.filename + '.sql')
        with open(file_path, 'r') as f:
            self.raw_sql = f.read()

    def find_values_to_replace(self):
        """Use pattern to identify all parameters in raw sql which need to be replaced."""
        regexp = re.compile(self.raw_pattern)
        self.to_replace = regexp.findall(self.raw_sql)

    def replace_params(self):
        """For every param, find and replace placeholder with appropriate value."""
        raw_sql = self.raw_sql
        for placeholder in self.to_replace:
            newreg = re.compile(placeholder)
            repl = self.get_replacement_value(placeholder)
            if repl:
                raw_sql = newreg.sub(str(repl), raw_sql)
        self.sql = raw_sql

    def get_replacement_value(self, to_be_replaced):
        """Given placeholder to be replaced, get name of parameter from w/in the pattern and lookup parameter value."""
        name_reg = re.compile('[a-zA-z_]+')
        param_name = name_reg.search(to_be_replaced).group()
        return self.params.get(param_name)

    def validate(self):
        """Check whether any placeholders have not been provided a replacement value."""
        regexp = re.compile(self.raw_pattern)
        missing = regexp.findall(self.sql)
        if missing:
            raise NameError("Value not provided for placeholder {}".format(missing))


class QueryExecutor(object):
    """
    QueryExecutor acts as a context manager for the execution of a one or more queries against the selected database.
    Keeps database connection open for duration of `with` block and closes afterwards.
    Usage:
    with QueryExecutor(db=MyDB) as qe:
        result = qe.execute_query(sql='select count(*) from my_table')
    """
    def __init__(self, db, logger=None):
        self.db = db
        self.cm = None
        self.logger = logger or PortholeLogger(name="QueryExecutor")

    def create_database_connection(self):
        self.cm = ConnectionManager(db=self.db, logger=self.logger)
        self.cm.connect()

    def close_database_connection(self):
        self.cm.close()

    def execute_query(self, filepath=None, filename=None, params=None, sql=None, multiple_statements=False):
        query = QueryGenerator(
            cm=self.cm,
            filepath=filepath,
            filename=filename,
            params=params,
            sql=sql,
            multiple_statements=multiple_statements,
            logger=self.logger
        )
        return query.execute()

    def commit(self):
        self.cm.commit()

    def __enter__(self):
        self.create_database_connection()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close_database_connection()


if __name__ == '__main__':
    pass



"""
Created on 03/23/2017.

Contains various helper functions related to time. By default,
dates and timestamps are returned as formatted strings. Just pass
string=False to change this behavior.

Sample Usage:

>>> import TimeHelper
>>> TimeHelper.current_timestamp()
'2017-03-23 22:50:31'
>>> TimeHelper.today()
'2017-03-23'
>>> TimeHelper.today(string=False)
datetime.date(2017, 3, 23)

"""

import datetime
import pytz


def current_timestamp(string: bool = True, timezone: str = 'UTC'):
    timestamp = datetime.datetime.now(tz=pytz.utc)
    if timezone != 'UTC':
        timestamp = convert_tz(timestamp, convert_to=timezone)
    if string:
        return timestamp_to_string(timestamp)
    else:
        return timestamp


def now(string: bool = True, timezone: str = 'UTC'):
    return current_timestamp(string, timezone)


def today(string: bool = True, timezone: str = 'UTC'):
    date = current_timestamp(string=False, timezone=timezone).date()
    if string:
        return date_to_string(date)
    else:
        return date


def timestamp_to_string(timestamp: datetime.datetime):
    return timestamp.strftime('%Y-%m-%d %-H:%M:%S')


def date_to_string(date: datetime.date):
    return date.strftime('%Y-%m-%d')


def convert_tz(timestamp: datetime.datetime, convert_to: str = 'US/Eastern'):
    return timestamp.astimezone(pytz.timezone(convert_to))


if __name__ == '__main__':
    pass



import sqlalchemy as sa
from porthole import ConnectionManager
from .logger import PortholeLogger
from porthole.models import automated_reports, automated_report_contacts, automated_report_recipients


class AutomatedReportContactManager(object):
    def __init__(self, database: str):
        self.database = database
        self.logger = PortholeLogger(name=__name__)

    def execute_statement(self, statement: sa.sql.base.Executable):
        with ConnectionManager(self.database) as cm:
            proxy = cm.conn.execute(statement)
            if proxy.cursor:
                result = proxy.fetchall()
            else:
                result = None
        return result

    def create_record(self, table: sa.Table, data: dict):
        self.execute_statement(table.insert().values(**data))

    def update_record(self, table: sa.Table, conditions: dict, data: dict):
        expression = self.compile_binary_expression(table, conditions)
        self.execute_statement(table.update().where(expression).values(**data))

    def delete_record(self, table: sa.Table, conditions: dict):
        expression = self.compile_binary_expression(table, conditions)
        self.execute_statement(table.delete().where(expression))

    @staticmethod
    def compile_binary_expression(table: sa.Table, field_value_dict: dict):
        expressions = [
            getattr(table.c, field) == value for field, value in field_value_dict.items()
        ]
        return sa.and_(*expressions)

    def record_exists(self, table: sa.Table, field_value_dict):
        expression = self.compile_binary_expression(table, field_value_dict)
        select = sa.sql.select([
            sa.exists().where(expression)
        ])
        try:
            return self.execute_statement(select)[0][0]
        except IndexError:
            return False

    def get_record_id(self, table: sa.Table, field, value):
        select = sa.sql.select(
            [table]
        ).where(
            getattr(table.c, field) == value
        )
        try:
            return self.execute_statement(select)[0][0]
        except IndexError:
            return None

    def report_exists(self, report_name):
        return self.record_exists(automated_reports, {'report_name': report_name})

    def add_report(self, report_name: str, active: int = 1):
        if self.report_exists(report_name):
            self.logger.warning(f"{report_name} already exists")
            return None
        self.create_record(
            table=automated_reports,
            data={
                'report_name': report_name,
                'active': active
            }
        )
        self.logger.info(f"Report '{report_name}' created successfully")

    def report_is_active(self, report_name):
        return self.record_exists(automated_reports, {'report_name': report_name, 'active': 1})

    def activate_report(self, report_name):
        if self.report_is_active(report_name):
            self.logger.warning(f"Report '{report_name}' is already active")
            return None
        self.update_record(automated_reports, {'report_name': report_name}, {'active': 1})
        self.logger.info(f"Report '{report_name}' is now active")

    def deactivate_report(self, report_name):
        if not self.report_is_active(report_name):
            self.logger.warning(f"Report '{report_name}' is already inactive")
            return None
        self.update_record(automated_reports, {'report_name': report_name}, {'active': 0})
        self.logger.info(f"Report '{report_name}' is now inactive")

    def contact_exists(self, email_address: str):
        return self.record_exists(automated_report_contacts, {'email_address': email_address})

    def add_contact(self, last_name: str = None, first_name: str = None, email_address: str = None):
        if self.contact_exists(email_address):
            self.logger.warning(f"Contact {last_name}, {first_name} ({email_address}) already exists ")
            return None
        self.create_record(
            table=automated_report_contacts,
            data={
                'last_name': last_name,
                'first_name': first_name,
                'email_address': email_address
            }
        )
        self.logger.info(f"Contact {last_name}, {first_name} ({email_address}) created successfully")

    def report_recipient_exists(self, report_name: str, email_address: str):
        return self.record_exists(
            table=automated_report_recipients,
            field_value_dict={
                'report_id': self.get_record_id(automated_reports, "report_name", report_name),
                'contact_id': self.get_record_id(automated_report_contacts, "email_address", email_address)
            }
        )

    def add_report_recipient(self, report_name: str, email_address: str, recipient_type: str):
        if recipient_type not in ['to', 'cc']:
            raise ValueError("Recipient type must be either `to` or `cc`.")
        if self.report_recipient_exists(report_name, email_address):
            self.logger.warning(f"Recipient '{email_address}' already exists for report '{report_name}'")
            return None
        self.create_record(
            table=automated_report_recipients,
            data={
                'report_id': self.get_record_id(automated_reports, 'report_name', report_name),
                'contact_id': self.get_record_id(automated_report_contacts, 'email_address', email_address),
                'recipient_type': recipient_type
            }
        )
        self.logger.info(f"{recipient_type} recipient '{email_address}' added successfully to report '{report_name}'")

    def remove_report_recipient(self, report_name: str, email_address: str):
        if not self.report_recipient_exists(report_name, email_address):
            self.logger.warning(f"Recipient '{email_address}' does not exist for report '{report_name}'")
            return None
        self.delete_record(
            table=automated_report_recipients,
            conditions={
                'report_id': self.get_record_id(automated_reports, 'report_name', report_name),
                'contact_id': self.get_record_id(automated_report_contacts, 'email_address', email_address),
            }
        )
        self.logger.info(f"Recipient '{email_address}' removed successfully from report '{report_name}'")



import os
from inspect import getfullargspec
from argparse import ArgumentParser
from .alerts import Alert
from .app import config
from .connections import ConnectionPool
from .mailer import Mailer
from .components import (DatabaseLogger,
                         RecipientsChecker,
                         ReportActiveChecker,
                         ReportErrorNotifier,
                         ReportWriter)
from .logger import PortholeLogger
from .uploaders import S3Uploader


class BasicReport:
    """
    A basic report with bare minimum functionality.

    Keyword arguments
    :report_title: Used in the filename of the resulting report.

    """
    def __init__(self, report_title, debug_mode=False, text_format='plain', logger_name=None, log_to_db=False):
        # -----------------------------------------
        # Assign arguments to instance attributes.
        # -----------------------------------------
        self.report_title = report_title
        self.report_writer = None
        self.logger = PortholeLogger(
            name=logger_name or report_title,
            log_to_db=log_to_db
        )
        self.to_recipients = []
        self.cc_recipients = []
        self.email = None
        self.subject = None
        self.message = None
        self.attachments = []
        self.debug_mode = debug_mode
        self.text_format = text_format
        self.email_sent = False
        self.failure_notification_sent = False
        self.file_path = config['Default'].get('base_file_path')
        self.default_db = config['Default'].get('database')
        self.conns = ConnectionPool(dbs=[self.default_db], logger=self.logger)

    def __del__(self):
        try:
            self.conns.close_all()
        except:
            pass

    @property
    def has_errors(self):
        return self.logger.error_buffer.present

    def get_conn(self, db):
        return self.conns.pool.get(db)

    def add_conn(self, db):
        return self.conns.add_connection(db)

    def build_file(self):
        report_writer = ReportWriter(report_title=self.report_title, logger=self.logger)
        report_writer.build_file()
        self.attachments.append(report_writer.report_file)
        self.report_writer = report_writer

    def create_worksheet_from_query(self,
                                    sheet_name,
                                    db=None,
                                    query=None,
                                    sql=None,
                                    query_kwargs=None,
                                    worksheet_kwargs=None):
        """Delegates functionality to ReportWriter."""
        if db is None:
            db = self.default_db
        cm = self.add_conn(db)
        self.report_writer.create_worksheet_from_query(
            cm=cm,
            sheet_name=sheet_name,
            query=query,
            sql=sql,
            query_kwargs=query_kwargs,
            worksheet_kwargs=worksheet_kwargs
        )

    def make_worksheet(self, sheet_name, query_results, **kwargs):
        """Delegates functionality to ReportWriter."""
        self.report_writer.make_worksheet(
            sheet_name=sheet_name,
            query_results=query_results,
            **kwargs
        )

    def build_email(self):
        """Instantiates Mailer object using provided parameters."""
        email = Mailer(
            recipients=self.to_recipients,
            cc_recipients=self.cc_recipients,
            debug_mode=self.debug_mode,
            text_format=self.text_format,
            logger=self.logger
        )
        email.subject = self.subject
        email.message = self.message
        email.attachments = self.attachments
        self.email = email

    def send_email(self):
        """Executes the send_email method of the Mailer."""
        try:
            self.email.send_email()
            self.email_sent = True
        except Exception as e:
            self.logger.exception(e)

    def check_whether_to_publish(self):
        """Determines whether email should be sent based given errors and settings."""
        if self.has_errors:
            return False
        else:
            return True

    def build_and_send_email(self):
        """If email should be sent, builds email object and sends email."""
        if self.check_whether_to_publish():
            self.build_email()
            self.send_email()

    def execute(self):
        """
        After instantiation and setup, and after
        any query results have been added to the workbook,
        this method executes and finalizes the report,
        ensures errors are handled, and performs cleanup.
        """
        if self.report_writer:
            self.report_writer.close_workbook()
        self.build_and_send_email()
        if self.has_errors:
            self.send_failure_notification()
        self.conns.close_all()

    def send_failure_notification(self):
        notifier = ReportErrorNotifier(
            report_title=self.report_title,
            error_buffer=self.logger.error_buffer.buffer
        )
        notifier.send_log_by_email()
        if notifier.notified:
            self.failure_notification_sent = True


class GenericReport(BasicReport):
    """A generic report object to be used to facilitate
    easy setup and configuration of new automated reports.

    Keyword arguments
    :report_title: Used in the filename of the resulting report.
    :report_name: Uniquely identifies a report. Used for recipient lookup
        and logging.
    :logging_enabled: (Optional) Default value of True. If logging is enabled,
        the report will attempt to create a log record each time it is instantiated.
        Only disable if the report needs to be run without connection to the standard
        reporting database.

    Here is an example of sample usage:

    report = GenericReport(
                            report_name='test_report',
                            report_title='Test Report'
                            )
    report.build_file()
    report.create_worksheet_from_query(
                                        sheet_name='Sheet1',
                                        query={'filename': 'test_query'}
                                        )
    report.subject = "Today's Test Report"
    report.message = "Attached is the latest report"
    report.execute()

    Notes on sample code:
    (1) A report is not required to have an attachment,
        but an attachment is assumed.
    (2) An Excel attachment can contain any number
        of worksheets. Simply call create_worksheet_from_query
        with the appropriate query parameters and worksheet
        name, as many times as you like. Worksheet names
        must be valid and unique.
    (3) After providing a subject and message body,
        the execute method will take care of the rest,
        including email circulation, error handling,
        and notification.

    Notes on database configuration:
    (1) Reports must have a record in the automated_reports table,
        including report_name and active flag (1 or 0).
    (2) Reports also must have at least one recipient
        in the automated_report_recipients table, and at least
        one recipient must be designed with a recipient type of
        'to'. Each report can have one or more recipients with
        a recipient type of 'cc'.
    (3) Report recipient records are associated with:
        (a) automated_reports through report_id
        (b) automated_report_contacts through contact_id
    (4) If you need a recipient who is not in the
        automated_report_contacts table, add them to it.

    """
    def __init__(
            self,
            report_title,
            report_name,
            logging_enabled=True,
            log_to_db=True,
            logger_name=None,
            send_if_blank=True,
            publish_to='email',
            debug_mode=False,
            text_format='plain'
    ):
        # -----------------------------------------
        # Assign arguments to instance attributes.
        # -----------------------------------------
        super().__init__(
            report_title=report_title,
            debug_mode=debug_mode,
            text_format=text_format,
            logger_name=logger_name or report_name,
            log_to_db=log_to_db
        )
        self.report_name = report_name
        self.logging_enabled = logging_enabled
        self.send_if_blank = send_if_blank
        if publish_to.lower() not in ['s3', 'email', 'both']:
            raise ValueError("Reports can only be published to s3 or email.")
        self.publish_to = publish_to.lower()
        self.db_logger = None
        self.active = None
        self.uploaded_to_s3 = False
        self.check_if_active()
        if self.active and self.logging_enabled:
            self.initialize_db_logger()

    @property
    def record_count(self):
        if self.report_writer:
            return self.report_writer.record_count
        else:
            return 0

    def initialize_db_logger(self):
        self.db_logger = DatabaseLogger(
            cm=self.get_conn(self.default_db),
            report_name=self.report_name,
            logger=self.logger
        )
        self.db_logger.create_record()

    def check_if_active(self):
        self.active = ReportActiveChecker(
            cm=self.get_conn(self.default_db),
            report_name=self.report_name,
            logger=self.logger
        )

    def get_recipients(self):
        """Performs lookup in database for report recipients based on report name."""

        checker = RecipientsChecker(
            cm=self.get_conn(self.default_db),
            report_name=self.report_name,
            logger=self.logger
        )
        self.to_recipients, self.cc_recipients = checker.get_recipients()

    def check_whether_to_publish(self):
        """Determines whether email should be sent based given errors, settings, and result count."""
        if self.has_errors:
            return False
        elif not self.send_if_blank and self.record_count == 0:
            return False
        else:
            return True

    def execute(self):
        """
        After instantiation and setup, and after
        any query results have been added to the workbook,
        this method executes and finalizes the report,
        ensures errors are handled, and performs cleanup.
        """
        if self.report_writer:
            self.report_writer.close_workbook()
        if self.active:
            if self.check_whether_to_publish():
                self.publish()
            if self.db_logger is not None:
                self.db_logger.finalize_record()
        if self.has_errors:
            self.send_failure_notification()
        self.conns.close_all()

    def publish(self):
        if self.publish_to == 'email':
            self.build_and_send_email()
        elif self.publish_to == 's3':
            self.upload_to_s3()
        elif self.publish_to == 'both':
            self.build_and_send_email()
            self.upload_to_s3()

    def upload_to_s3(self):
        try:
            uploader = S3Uploader(debug_mode=self.debug_mode)
            uploaded = []
            for attachment in self.attachments:
                success = uploader.upload_file(key=os.path.basename(attachment), filename=attachment)
                uploaded.append(success)
            self.uploaded_to_s3 = all(uploaded)
        except:
            self.logger.exception("Unable to upload to S3")


class ReportRunner(ArgumentParser):
    def __init__(self, report_map=None):
        super().__init__(description="Runs the report designated by the -r or --report parameter. To see a list of available reports, use -l or --list.")
        self.add_argument("-l", "--list", action='store_true', dest='list', help="show a list of available reports")
        self.add_argument("-r", "--report", dest='report', help="name of report to run")
        self.add_argument("-d", "--debug", action='store_true', dest='debug_mode', help="run the requested report in debug mode, if defined")
        self.add_argument("-p", "--ping", action='store_true', dest='ping', help="this is used for health checking.")
        self.args, _ = super().parse_known_args()
        self.report_map = report_map if report_map is not None else {}
        self.logger = PortholeLogger("report_runner", log_to_db=True)

    def handle_args(self):
        if self.args.list:
            self.list_reports()
        elif self.args.report:
            self.run_report()

    def list_reports(self):
        print("The following reports are available to run:")
        for report in self.report_map.keys():
            print(report)

    def run_report(self):
        report_name = self.args.report
        report_function = self.report_map.get(report_name)
        if report_function:
            if 'debug_mode' in getfullargspec(report_function).args:
                self.logger.info(f"Received call to run {report_name} in debug mode.")
                report_function(debug_mode=self.args.debug)
            else:
                self.logger.info(f"Received call to run {report_name}")
                report_function()
        else:
            self.send_unmapped_report_alert()
            error_message = "Report Runner is unable to run: {}. There is no report function mapped to this name".format(report_name)
            self.logger.error(error_message)
            raise ValueError(error_message)

    def send_unmapped_report_alert(self):
        alert = Alert(
            subject=f"Attempt to execute unmapped report <{self.args.report}>",
            message="Check to ensure report mapping is correct.",
            recipients=[config['Admin']['admin_email']]
        )
        alert.send()


if __name__ == '__main__':
    pass



from porthole import config
from porthole.connections import ConnectionPool
from porthole.components import (
    DatabaseLogger,
    PortholeLogger,
    ReportActiveChecker
)


class DataTask:
    """
    DataTask is meant to be used when you would like to run some arbitrary function with basic Porthole-style
    logging and error handling. The task name must be defined in the `automated_reports` table.
    If the callable task function executes without raising an exception, successful execution will be recorded.
    If an exception is raised by the task function, a failed execution will be recorded along with the exception. The
    exception will be raised, and should therefore be handled in any application code.
    """
    def __init__(self, task_name, task_function, logging_enabled=True, log_to_db=False):
        if not callable(task_function):
            raise TypeError(
                f"DataTask <{task_name}> requires a callable task function, but <{task_function}> is not callable."
            )
        self.task_name = task_name
        self.task_function = task_function
        self.logging_enabled = logging_enabled
        self.logger = PortholeLogger(
            task_name,
            log_to_db=log_to_db
        )
        self.default_db = config['Default'].get('database')
        self.conns = ConnectionPool(dbs=[self.default_db], logger=self.logger)
        self.db_logger = None
        self.active = None
        self.success = None
        self.check_if_active()
        if self.active and self.logging_enabled:
            self.initialize_db_logger()

    def get_conn(self, db):
        return self.conns.pool.get(db)

    def add_conn(self, db):
        return self.conns.add_connection(db)

    def initialize_db_logger(self):
        self.db_logger = DatabaseLogger(
            cm=self.get_conn(self.default_db),
            report_name=self.task_name,
            logger=self.logger
        )
        self.db_logger.create_record()

    def check_if_active(self):
        self.active = ReportActiveChecker(
            cm=self.get_conn(self.default_db),
            report_name=self.task_name,
            logger=self.logger
        )

    def execute(self):
        err = None
        try:
            if self.active:
                self.task_function()
                self.success = True
        except Exception as e:
            err = e
            self.success = False
            self.logger.exception(e)
        finally:
            if self.db_logger is not None:
                self.db_logger.finalize_record()
            self.conns.close_all()
            if err is not None:
                raise err



import datetime
import xlsxwriter
import openpyxl


class WorkbookBuilder(object):
    """
    Allows for creation of Excel workbooks containing arbitrary worksheets
    using the xlsxwriter library.
    """

    DEFAULT_COL_WIDTH = 10
    MAX_COLUMN_WIDTH = 50
    DEFAULT_HEADER_PARAMS = {'bold': True}

    def __init__(self, filename=None):
        self.filename = filename
        self.default_header_format = None
        self.formats = {}
        self.workbook_options = {
            'constant_memory': True,
            'default_date_format': 'mm/dd/yy'}
        self.workbook = None
        if filename:
            self.create_workbook()

    def create_workbook(self):
        """Given filename, workbook options, and head format, create workbook."""
        workbook = xlsxwriter.Workbook(self.filename, self.workbook_options)
        self.default_header_format = workbook.add_format(WorkbookBuilder.DEFAULT_HEADER_PARAMS)
        self.workbook = workbook

    def close_workbook(self):
        """An exception may be raised if workbook is not closed explicitly."""
        self.workbook.close()

    def add_format(self, format_name, format_params):
        """Creates new format object and adds it to this workbook's available formats."""
        if format_name in self.formats:
            raise IndexError(f"Workbook already contains format object < {format_name} >")
        self.formats[format_name] = self.workbook.add_format(format_params)

    def add_worksheet(
            self, sheet_name, field_names, sheet_data,
            format_axis=None, format_rules=None, row_start=0, col_start=0,
            autofit_columns=True, column_width=None, freeze_first_row=False,
            header_format=None, show_autofilter=False
    ):
        """
        :param sheet_name: Worksheet name as string.
        :param field_names: List of column names to be written as headers.
        :param sheet_data: List of lists containing data to write to the worksheet.
        :param format_axis: Optional, default None. Allowed values include `0` or `'row'`, or `1` or `column`.
        :param format_rules: Optional, default None. Dictionary of parameters for format rules to be applied.
        :param row_start: Optional, default value of 0 (first row). Set the row to start writing from.
        :param col_start: Optional, default value of 0 (first column). Set the column to start writing from.
        :param autofit_columns: Optional, default value of False. Select whether columns should be "auto-fitted".
        :param column_width: Optional, default of None. Specify a uniform column width.
        :param freeze_first_row: Optional, default False. Freezes first row when scrolling.
        :param header_format: Optional, default None. Apply specified format name to header row.
        :param show_autofilter: Optional, default False. Show autofilter on first row.

        Add worksheet to the workbook. This method assumes that each inner list (row) has the same number of elements.
        """
        if autofit_columns and column_width:
            raise ValueError("Cannot both auto-fit columns and set a uniform column width.")
        worksheet = self.workbook.add_worksheet(sheet_name)
        cell_formats = create_cell_formats(sheet_data, format_axis, format_rules)

        e_row = row_start
        e_col = col_start

        # Write field names in first row.
        for i, field_name in enumerate(field_names):
            worksheet.write(
                e_row,
                e_col + i,
                field_name,
                self.formats.get(header_format, self.default_header_format)
            )

        # Increment row counter for incoming data.
        e_row += 1

        # Write the data.
        for value_row, format_row in zip(sheet_data, cell_formats):
            for x, (val, fmt) in enumerate(zip(value_row, format_row)):
                worksheet.write(e_row, e_col + x, val, self.formats.get(fmt))
            e_row += 1

        col_widths = self.calculate_column_widths(field_names, sheet_data, autofit_columns, column_width)
        for i, width in enumerate(col_widths):
            worksheet.set_column(i, i, width)
        if freeze_first_row:
            worksheet.freeze_panes(1, 0)
        if show_autofilter:
            worksheet.autofilter(0, 0, 0, len(field_names) - 1)

    def calculate_column_widths(self, field_names, sheet_data, autofit_columns=False, column_width=None):
        """
        :param field_names:
        :param sheet_data:
        :param autofit_columns: Optional, default value of False. Select whether columns should be "autofitted".
        :param column_width: Optional, default of None. Specify a uniform column width.
        :return: List of column widths as integers
        """
        column_width = column_width or WorkbookBuilder.DEFAULT_COL_WIDTH
        col_widths = [column_width] * len(field_names)

        if autofit_columns is True:
            # We want the longest value in a given column, including the field name.
            headers_and_data = [field_names] + sheet_data
            for row in headers_and_data:
                for idx, value in enumerate(row):
                    if isinstance(value, datetime.datetime):
                        # datetime string representation is usually much longer (26 chars)
                        # than default date format (8 chars)
                        formatted_length = len(str(self.workbook.default_date_format.num_format))
                        compare_value = max(formatted_length, WorkbookBuilder.DEFAULT_COL_WIDTH)
                    else:
                        compare_value = len(str(value))
                    if compare_value > col_widths[idx]:
                        col_widths[idx] = compare_value * 1.15

        for i, width in enumerate(col_widths):
            if width > WorkbookBuilder.MAX_COLUMN_WIDTH:
                col_widths[i] = WorkbookBuilder.MAX_COLUMN_WIDTH

        return col_widths


class WorkbookEditor(object):
    def __init__(self, workbook_filename):
        self.workbook_filename = workbook_filename
        self.workbook = openpyxl.load_workbook(filename=workbook_filename)

    def replace_sheet_contents(self, sheet_name: str, data_rows: list, headers: list) -> None:
        del self.workbook[sheet_name]
        worksheet = self.workbook.create_sheet(sheet_name)
        worksheet.append(headers)
        for row in data_rows:
            worksheet.append(list(row))

    def get_all_values(self, sheet_name: str) -> list:
        rows = []
        for row in self.workbook[sheet_name].values:
            rows.append(row)
        return rows
    
    def save_workbook(self, save_as: str=None) -> None:
        if save_as is None:
            save_as=self.workbook_filename
        self.workbook.save(save_as)


def apply_column_rule(row, rule):
    """
    Given a cell value and a rule, return a format name. See `create_cell_formats` for more detail.
    """
    rule_type = rule['type']
    if rule_type == 'all':
        return rule['format']
    if rule_type == 'boolean':
        if rule['function'](row):
            return rule['format']
        else:
            return None
    if rule_type == 'conditional':
        return rule['function'](row)
    raise ValueError(f"Column rule type <{rule['type']}> not supported")


def create_column_formats(data, format_rules):
    """
    Iterate through provided data and apply format rules, returning format names corresponding to each data element.
    """
    result = []
    for row in data:
        row_formats = [
            apply_column_rule(row, format_rules[col_name]) if col_name in format_rules else None
            for col_name in row.keys()
        ]
        result.append(row_formats)
    return result


def create_row_formats(data, format_rules):
    raise NotImplementedError("Formatting rows has not yet been implemented, but will be included in a future release.")


def create_cell_formats(data, format_axis=None, format_rules=None):
    """
    Apply format rules (if provided) to data and return format names corresponding to each data element. Format rules
    should contain the names of a specific format object, not the format object itself or paramteters to construct one.
    All format names referenced should be added separately using the `WorkbookBuilder.add_format` method.

    :param data: The results of a query, or other data structure, to be written to a new worksheet.
    :type data: List of RowDict objects, e.g. `QueryResult.result_data`
    :param format_axis: The axis along which to apply the provided format rules - formats can be applied to rows,
        or to columns, but not both.
    :type format_axis: Integer, string, or None.
    :param format_rules: The rule(s) to apply to the data. Available format and options are as follows:

    Columns - Keys are field names as strings. Values are dictionaries representing the rule to apply for that specific
        column. These dictionaries must contain `type`, and may contain `format` and/or `function`.

    Possible values for `type` include:
        all: format the entire column using the provided format name.
        boolean: format any values with the provided format name, which test as True using a provided `function`,
            The functions should accept a RowDict and return True or False.
        conditional: format any values with one or more provided format names. The provided `function` should accept
            cell values and return the format name (or None).

    Sample usage:

    format_rules = {
        'salary': {'type': 'all', 'format': 'money'},
        'age': {'type': 'boolean', 'format': 'bold_red', 'function': lambda row: row['field'] % 2 == 0},
    }

    :type format_rules: Dictionary, with keys representing column names (if applying column rules).
    :return: List of lists of strings, where each is the name of the format object to be applied to that data
        element, or None if no format is to be applied.
    """
    if format_axis not in (None, 0, 1, 'row', 'column'):
        raise ValueError("Must provide valid format axis: 0 or 'row', 1 or 'column'.")
    if format_axis in (1, 'column'):
        return create_column_formats(data, format_rules)
    if format_axis in (0, 'row'):
        return create_row_formats(data, format_rules)
    return [[None] * len(row) for row in data]



try:
    import boto3
    from boto3.s3.transfer import TransferConfig
except ImportError:
    boto3 = None
    TransferConfig = None
from .app import config
from .logger import PortholeLogger


class S3Uploader(object):
    def __init__(self, debug_mode: bool = False) -> None:
        if boto3 is None or TransferConfig is None:
            raise ModuleNotFoundError(
                "boto3 is a required dependency to use the S3Uploader class, but is not currently installed."
            )
        self.logger = PortholeLogger("porthole.Uploaders")
        self.debug_mode = debug_mode or config.getboolean('Debug', 'debug_mode')
        s3_profile = config['AWS']['S3']
        session_params = {
            'aws_access_key_id': config[s3_profile]['aws_access_key_id'],
            'aws_secret_access_key': config[s3_profile]['aws_secret_access_key'],
            'region_name': config[s3_profile]['region_name']
        }
        self.bucket = config[s3_profile]['bucket']
        session = boto3.Session(**session_params)
        self.session = session.client('s3')

    def upload_file(self, key: str, filename: str, bucket: str = None, s3config: TransferConfig = None) -> bool:
        if self.debug_mode:
            self.logger.info(
                f"Debug mode active. Would have uploaded {filename} to {bucket}/{key}."
            )
            return True
        else:
            return self._upload_file(key, filename, bucket, s3config)

    def _upload_file(self, key: str, filename: str, bucket: str = None, s3config: TransferConfig = None) -> bool:
        if bucket is None:
            bucket = self.bucket
        if s3config is None:
            s3config = TransferConfig(multipart_threshold=8388608 * 6)
        try:
            self.session.upload_file(Bucket=bucket, Key=key, Filename=filename, Config=s3config)
            self.logger.info(f"Uploaded {filename} to {bucket}/{key}")
            return True
        except FileNotFoundError:
            self.logger.exception(f"Unable to upload {filename}. File not found.")
            return False
        except Exception:
            self.logger.exception("Exception occurred during upload.")
            raise



from setuptools import setup

setup(
    name='porthole',
    version='0.8.2',
    description='An automated reporting package.',
    author='Billy McMonagle',
    author_email='speedyturkey@gmail.com',
    url='https://github.com/speedyturkey/porthole',
    packages=['porthole'],
    python_requires='>3.6',
    install_requires=[
        'openpyxl',
        'psycopg2-binary',
        'pymysql',
        'pytz',
        'SQLAlchemy',
        'xlsxwriter',
    ],
    extras_require={
        'AWS': ["boto3"]
    },
    zip_safe=False
)



import os
from configparser import ConfigParser


class PortholeConfig(ConfigParser):
    CONFIG_ENV_NAME = 'PORTHOLE_CONFIG'
    DEFAULT_CONFIG_FILE = 'porthole.ini'
    DEFAULT_CONFIG_DIRECTORY = 'config'
    OTHER_ALLOWED_CONFIG_PATHS = [
        os.path.join(DEFAULT_CONFIG_DIRECTORY, 'porthole.ini'),
        os.path.join(DEFAULT_CONFIG_DIRECTORY, 'config.ini')
    ]

    def __init__(self) -> None:
        super().__init__(os.environ, interpolation=None)
        config_file = self.determine_config()
        self.read(config_file)

    @staticmethod
    def determine_config() -> str:
        """
        Porthole needs to identify which config file should be used. A cascading approach is utilized that is flexible
        but has simple default options. Either set the environment variable PORTHOLE_CONFIG,
        or review the PortholeConfig docstrings.
        """
        if os.environ.get(PortholeConfig.CONFIG_ENV_NAME) is not None:
            return os.environ.get(PortholeConfig.CONFIG_ENV_NAME)
        if os.path.isfile(PortholeConfig.DEFAULT_CONFIG_FILE):
            return PortholeConfig.DEFAULT_CONFIG_FILE
        for file_path in PortholeConfig.OTHER_ALLOWED_CONFIG_PATHS:
            if os.path.isfile(file_path):
                return file_path
        raise FileNotFoundError(
            "Porthole is unable to locate a useable config file. "
            "Try setting the PORTHOLE_CONFIG environment variable, "
            "or creating a porthole.ini file in your main project directory."
        )


config = PortholeConfig()

if __name__ == '__main__':
    pass



import sys
try:
    from .app import config
    from .connections import ConnectionManager
    from .contact_management import AutomatedReportContactManager
    from .getting_started import new_config, setup_tables
    from .filters import ResultFilter
    from .logger import PortholeLogger
    from .mailer import Mailer
    from .related_record import RelatedRecord
    from .reports import BasicReport, GenericReport, ReportRunner
    from .tasks import DataTask
    from .queries import QueryExecutor, QueryGenerator, QueryReader, QueryResult
    from .workflows import SimpleWorkflow
    from .xlsx import WorkbookBuilder, WorkbookEditor
except KeyError:
    print("Unable to import Porthole due to KeyError. Check config/config.ini.")
    print("{}: {}".format(sys.exc_info()[1].__doc__, sys.exc_info()[1]))
